#!/usr/bin/env python3

# This script is part #6 to reproduce the results of Bletery and Nocquet (Science, 2023)
# It simulates the approach devlopped in part #1-5 many times in order to evaluate how frequently the signal we hilghlight could randomly arise from noise

from glob import glob
import numpy as np
import pyacs.lib.astrotime as at
from datetime import datetime
import os
import matplotlib.pyplot as plt
import shutil
from random import randrange, uniform

safe_time = 0   
nsample = 576
tlapse = 6 # time window increment (in hours)
st_threshold = 75 # minimum percentage of stations with complete time series to consider a test time window
threshold = 10 # ad-hoc threshold to detect offsets in stacks 
nd = int(nsample/2)
w = 12 # time window to find offset (in samples)

test_dir = 'test_directory'
data_dir = 'data'
data_byEQ_dir = 'data_byEQ'
leq_file = glob(data_byEQ_dir+"/20*_MW_*/dot_product_stack.txt")

if os.path.exists(test_dir):
    shutil.rmtree('./'+test_dir)
os.makedirs(test_dir)

for eq_file in leq_file:

    print("Processing EQ %s" % eq_file.split('/')[-2])
    
    eq_test_dir = test_dir+'/'+eq_file.split('/')[-2]
    if not os.path.exists(eq_test_dir):
        os.makedirs(eq_test_dir)

    date_eq_as_datetime = datetime.strptime(eq_file.split('/')[-2].split('.')[0].split('_')[0], '%Y%m%dT%H:%M:%S')
    date_eq_seconds = at.datetime2seconds(date_eq_as_datetime)    
    mjd_eq = at.datetime2mjd(date_eq_as_datetime)
    (np_doy,np_year,np_ut) = at.mjd2dayno([mjd_eq-2,mjd_eq-1,mjd_eq])
    sec_J2000 = int(at.decyear2seconds(2000.0,rounding='second')+ 12*3600) 
    
    eq_dir = data_byEQ_dir+'/'+eq_file.split('/')[-2]
    np_code = np.atleast_1d(np.genfromtxt(eq_dir+'/station_list.txt',dtype=str,usecols=0))
    syn_east = np.atleast_1d(np.genfromtxt(eq_dir+'/synthetics.txt',dtype=float,usecols=0))
    syn_north = np.atleast_1d(np.genfromtxt(eq_dir+'/synthetics.txt',dtype=float,usecols=1))
    dot_product0, dot_product = np.zeros([len(np_code), nsample]), np.zeros([len(np_code), nsample])
    c = 0
    
    days_to_avoid = list(range(int(np_doy[0]),int(np_doy[0])+92)) # avoid 2 days before earthquakes and 90 days after (to limit post-seismic signal)

    for i in range(3,366):
        if not i in days_to_avoid:
            for j in range(0,24,tlapse):
                fake_time = uniform(0,tlapse)+j 
                date_fake_eq_as_datetime = at.dayno2datetime(i,int(np_year[2]),ut=fake_time/24)
                date_fake_eq_sec = at.datetime2seconds(date_fake_eq_as_datetime)
                date_fake_eq_sec_J2000 = date_fake_eq_sec - sec_J2000
                N = 0
                sigmag2, sigmaj2 = [], []
                for code in np_code:
                    k = np.where(np_code==code)[0][0]
                    try:
                        ts0 = np.genfromtxt((data_dir+"/%s.%04d.%03d.kenv.gz" % (code,np_year[0],i-2)), skip_header=1)
                        ts1 = np.genfromtxt((data_dir+"/%s.%04d.%03d.kenv.gz" % (code,np_year[1],i-1)), skip_header=1)
                        ts2 = np.genfromtxt((data_dir+"/%s.%04d.%03d.kenv.gz" % (code,np_year[2],i)), skip_header=1)
                    except:
                        continue
                    ts = np.concatenate((ts0,ts1,ts2),axis=0)[:,1:11]
                    np_sec_J2000 = ts[:,0].astype(int)    
                    np_sec_wrt_eq = np_sec_J2000 - date_fake_eq_sec_J2000
                    idx = np.where((np_sec_wrt_eq <= 0))[0][-1] 
                    if np_sec_wrt_eq[idx] < -300: 
                        continue
                    try:
                        sample_date = np_sec_J2000[idx-nsample:idx]
                        maxsamp = np.max(np.diff(sample_date))
                    except:
                        continue
                    if maxsamp > 300: 
                        continue
                    if len(sample_date) < nsample: 
                        continue 
                    east, north = ts[idx-nsample:idx,7], ts[idx-nsample:idx,8] 
                    east0, north0 = np.median(east[0:nd]), np.median(north[0:nd])
                    norme, normn = np.linalg.norm(east[0:nd]-east0), np.linalg.norm(north[0:nd]-north0)
                    sj2 = norme*norme + normn*normn
                    dot_product0[k,:] = (east-east0)*syn_east[k] + (north-north0)*syn_north[k]  
                    dot_product[k,:] = dot_product0[k,:] / sj2
                    sg2 = syn_east[k]*syn_east[k] + syn_north[k]*syn_north[k]
                    sigmag2.append(float(sg2))
                    sigmaj2.append(float(sj2))
                    N = N+1

                if N >= (st_threshold/100) * len(np_code):      

                    # Tests if there is an offset     
                    eq_stack0 = sum(dot_product0) # (easier to see before the normalization by sj2)
                    d_eq_stack0 = abs(np.diff(eq_stack0))
                    I = np.where(d_eq_stack0==max(d_eq_stack0[w-1:-w-1]))[0][0]
                    offset = abs(np.mean(eq_stack0[I+1:I+w+1]) - np.mean(eq_stack0[I-w+1:I+1])) / np.std(eq_stack0[I-w+1:I+1])
                    if offset > threshold:
                        continue
        
                    eq_stack = sum(dot_product)
                    fake_hour = int(fake_time)
                    fake_min = int((fake_time-fake_hour)*60)
                    eqt_dir = eq_test_dir+"/%04d.%03d.%02d.%02d" % (np_year[0],i,fake_hour,fake_min)
                    if os.path.exists(eqt_dir):
                        shutil.rmtree(eqt_dir)
                    os.makedirs(eqt_dir)
                    np.savetxt(eqt_dir+'/dot_product_stack.txt',eq_stack)
                    np.savetxt(eqt_dir+'/sigmag2.txt',sigmag2)
                    np.savetxt(eqt_dir+'/sigmaj2.txt',sigmaj2)
                    c = c+1
    print(str(c)+' test-stacks made\n')

